#!/usr/bin/env python
# -*- coding: utf-8 -*-
import argparse
import json
import os
import sys
import time
import urllib
from datetime import datetime
from urllib.request import Request, urlopen
from uuid import UUID


SRC_CURRENCIES = ['btc', 'eth', 'dash', 'zec', 'xmr', 'ltc', 'kmd', 'bdx', 'oxen', 'firo']


def get_http_address(onion):
    if onion:
        return 'http://xmxmrjoqo63c5notr2ds2t3pdpsg4ysqqe6e6uu2pycecmjs4ekzpmyd.onion'
    return 'https://xchange.me'


def get_api_endpoint(onion, endpoint, data = None):
    url = get_http_address(onion) + "/api/v1/" + endpoint
    if data is not None:
        url += '?' + urllib.parse.urlencode(data)
    return url


def validate_uuid(uuid):
    try:
        UUID(uuid)
    except Exception:
        msg = "%s is not a valid UUID" % uuid
        raise argparse.ArgumentTypeError(msg)
    return uuid


def make_request(req):
    try:
        resp = urllib.request.urlopen(req)
        body = resp.read().decode('utf8')
        data = json.loads(body)
        resp.close()
    except urllib.error.HTTPError as e:
        if e.code == 422:
            print('Following errors have occurred:')
            error = json.loads(e.read().decode('utf8'))
            for key in error['errors']:
                print(key + ": " + error['errors'][key][0])
        elif e.code == 404:
            print('Requested data cannot be found.')
        else:
            raise e
        exit(-5)
    else:
        return data


def check_rate(args):
    data = {
        'from_currency': args.from_currency,
        'to_currency': args.to_currency,
    }

    if hasattr(args, 'amount'):
        data['amount'] = args.amount

    req = Request(get_api_endpoint(args.onion, "exchange/estimate", data), headers={
        'Accept': 'application/json', 'User-Agent': 'xchangeme-cli',
    })
    estimate = make_request(req)
    print('1.0 %s ~= %s %s' % (args.from_currency, estimate['rate'], args.to_currency))
    if estimate['estimate'] != 0:
        print('%s %s ~= %s %s' % (args.amount, args.from_currency, estimate['estimate'], args.to_currency))


def create_exchange(args):
    data = {
        'receive_address': args.receive_address,
        'extra_id': args.extra_id,
        'from_currency': args.from_currency,
        'to_currency': args.to_currency,
        'withdraw_to': args.withdraw_to,
        'from_currency_alt': args.z_address,
        'from_amount': args.amount,
        'refund_address': args.refund_address,
        'invite_code': args.invite_code,
    }
    req = Request(get_api_endpoint(args.onion, 'exchange'), headers={
        'Accept': 'application/json',
        'Content-Type': 'application/json',
        'User-Agent': 'xchangeme-cli',
    }, data=json.dumps(data).encode(), method='POST')
    exchange = make_request(req)
    if exchange['from_amount'] == 0:
        amount = 'at least %s' % exchange['minimum_payment']
    else:
        amount = str(exchange['from_amount'])
    print('New exchange has been created.')
    print('Please pay %s %s to %s' % (amount, exchange['from_currency'], exchange['send_address']))
    print('We will be waiting for your payment for the next 5 hours.')
    print('You may check your exchange status by visiting %s or calling %s check-status %s' %
          (get_http_address(args.onion) + '/exchange/' + exchange['uuid'], sys.argv[0], exchange['uuid']))
    if args.show_pgp_proof:
        print(exchange['pgp_proof'])
    else:
        print('PGP proof is available on %s/%s/proof' % (get_http_address(args.onion), exchange['uuid']))


def check_status(args):
    exchange = make_request(Request(get_api_endpoint(args.onion, "exchange/" + args.uuid), headers={
     'User-Agent': 'xchangeme-cli', 'Accept': 'application/json',}))
    if exchange['stage'] == 'UNPAID':
        print('We are looking on blockchain for your payment.')
        print('If you did not pay yet, please make a payment to the address %s.' % exchange['send_address'])
        print('If you already paid, please wait for the required number of confirmations before proceeding.')
        print('We will be waiting for your payment until %s UTC.' %
              datetime.fromtimestamp(time.time() + exchange['time_left_for_payment']))
    elif exchange['stage'] == 'DONE':
        print('Your exchange has been completed.')
        print('We have sent you %s to the address %s.' % (exchange['send_amount'], exchange['receive_address']))
    elif exchange['stage'] == 'REFUNDED':
                print('Your exchange has been refunded.')
                print('We have sent you %s minus miners fee, to the address %s.' % (exchange['from_amount'], exchange['refund_address']))
    elif exchange['stage'] == 'REFUND_FAILED':
                    print('Your exchange cannot be refunded as you have never provided refund address.')
                    print('Please contact support and provide refund address.')
    else:
        print('We are currently processing your exchange.')
        print('Please wait for the process to finish.')
    print('---')
    print('More details are available on www (JavaScript NOT required), please visit %s'
          % get_http_address(args.onion) + '/exchange/' + args.uuid)
    print('If you have any problem with your exchange or any questions, send us an email at support@xchange.me')


def show_pgp_proof(args):
    exchange = make_request(Request(get_api_endpoint(args.onion, "exchange/" + args.uuid), headers={
    'User-Agent': 'xchangeme-cli', 'Accept': 'application/json',}))
    print(exchange['pgp_proof'])


def list_destinations(args):
    destinations = make_request(Request(get_api_endpoint(args.onion, 'currencies/to'), headers={
    'User-Agent': 'xchangeme-cli', 'Accept': 'application/json',}))
    print('Available destination currencies: ' + ', '.join(destinations) + '.')


def is_running_on_onion_os():
    if os.path.exists('/usr/share/anon-ws-base-files'):
        return 'Whonix Workstation'
    try:
        with open('/etc/os-release') as f:
            if 'TAILS_PRODUCT_NAME="Tails"' in f.read():
                return 'Tails'
    except FileNotFoundError:
        return False
    return False


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog='xchangeme-cli.py',
                                     description='CLI utility for xchange.me, the crypto exchanging service',
                                     epilog='visit us on https://xchange.me or http://xmxmrjoqo63c5notr2ds2t3pdpsg4ysqqe6e6uu2pycecmjs4ekzpmyd.onion')
    parser.add_argument('--onion', action='store_true', default=False,
                        help='connect to onion site instead of clear-net one')
    parser.add_argument('--skip-onion-os-check', default=False, action='store_true',
                        help='skip checking if running on Tails or Whonix Workstation')
    subparsers = parser.add_subparsers(description='select one of following operations:')

    check_rate_parser = subparsers.add_parser('check-rate', description='check current exchange rate')
    check_rate_parser.set_defaults(func=check_rate)
    check_rate_parser.add_argument('from_currency', help='source currency (currency of coins you send)',
                                   choices=SRC_CURRENCIES)
    check_rate_parser.add_argument('to_currency', help='destination currency (currency of coins you will receive)')
    check_rate_parser.add_argument('--amount', type=float, help='how much of coins you want to send (source currency)')

    create_exchange_parser = subparsers.add_parser('create-exchange', description='Create new exchange')
    create_exchange_parser.set_defaults(func=create_exchange)
    create_exchange_parser.add_argument('from_currency', help='source currency (currency of coins you send)',
                                        choices=SRC_CURRENCIES)
    create_exchange_parser.add_argument('to_currency', help='destination currency (currency of coins you will receive)')
    create_exchange_parser.add_argument('receive_address',
                                        help='destination address (address on which you will receive coins)')
    create_exchange_parser.add_argument('--extra-id', default=None,
                                        help='extra ID if needed (i.e. Destination TAG for xrp)')
    create_exchange_parser.add_argument('--withdraw-to', default='eth',
                                        help='select destination chain if coin exists on multiple chains '
                                             '(i.e. for usdt you can select eth, omni or trx and eth is default)')
    create_exchange_parser.add_argument('--z-address', default=False, action='store_true',
                                        help='enables private (z) address for zec')
    create_exchange_parser.add_argument('--amount', default=None, type=float,
                                        help='amount you want to send (used only for estimation and validation)')
    create_exchange_parser.add_argument('--show-pgp-proof', default=False, action='store_true', help='show PGP proof')
    create_exchange_parser.add_argument('--refund-address', default=None, help='refund address')
    create_exchange_parser.add_argument('--invite-code', default=None,
                                        help='invite code from affiliate program (taken from affiliate dashboard)')

    check_status_parser = subparsers.add_parser('check-status', description='Check exchange status')
    check_status_parser.set_defaults(func=check_status)
    check_status_parser.add_argument('uuid', help='Unique ID of exchange you want to check', type=validate_uuid)

    show_pgp_proof_parser = subparsers.add_parser('show-pgp-proof', description='Check exchange PGP proof')
    show_pgp_proof_parser.set_defaults(func=show_pgp_proof)
    show_pgp_proof_parser.add_argument('uuid', help='Unique ID of exchange you want to check', type=validate_uuid)

    list_destinations_parser = subparsers.add_parser('list-destinations',
                                                     description='List currently available destination currencies')
    list_destinations_parser.set_defaults(func=list_destinations)

    parser_args = parser.parse_args()

    if not hasattr(parser_args, 'func'):
        parser.print_help(sys.stderr)
        exit(1)

    if parser_args.onion is False and parser_args.skip_onion_os_check is False:
        os = is_running_on_onion_os()
        if os is not False:
            print('%s detected, automatically switching to onion endpoint.' % os)
            parser_args.onion = True

    parser_args.func(parser_args)
